@@ -39,21 +39,20 @@
 #include <libgnomeprint/private/gnome-print-private.h>
 #include <libgnomeprint/gnome-print-meta.h>
 
-#include "gnome-print-preview-private.h"
-#include "gnome-print-i18n.h"
-#include "gnome-print-job-preview.h"
+#include <libgnomeprintui/gnome-print-i18n.h>
+#include <libgnomeprintui/gnome-print-job-preview.h>
+
+#define GPP_COLOR_RGBA(color, ALPHA) \
+	((guint32) (ALPHA              | \
+	 (((color).red   / 256) << 24) | \
+	 (((color).green / 256) << 16) | \
+	 (((color).blue  / 256) <<  8)))
 
 #define GPMP_ZOOM_IN_FACTOR M_SQRT2
 #define GPMP_ZOOM_OUT_FACTOR M_SQRT1_2
 #define GPMP_ZOOM_MIN 0.0625
 #define GPMP_ZOOM_MAX 16.0
 
-#define GPMP_A4_WIDTH (210.0 * 72.0 / 25.4)
-#define GPMP_A4_HEIGHT (297.0 * 72.0 / 2.54)
-
-#define TOOLBAR_BUTTON_BASE 5
-#define MOVE_INDEX 5
-
 #ifndef GTK_STOCK_EDIT
 #define GTK_STOCK_EDIT "gtk-edit"
 #endif
@@ -172,9 +171,30 @@ struct _GnomePrintJobPreviewClass {
 	GtkWindowClass parent_class;
 };
 
+static void gnome_print_job_preview_get_targets (GnomePrintJobPreview *jp,
+						 GtkTargetEntry      **t,
+						 guint                *nt);
+static void target_entries_free (GtkTargetEntry *t, guint nt);
+
 static void
 gnome_print_job_preview_close (GnomePrintJobPreview *jp)
 {
+	g_return_if_fail (GNOME_IS_PRINT_JOB_PREVIEW (jp));
+
+	/* Store the clipboard before we quit. */
+	if (jp->clipboard) {
+		GdkDisplay *d;
+		GtkClipboard *c;
+		GtkTargetEntry *t = NULL;
+		guint nt = 0;
+
+		d = gtk_widget_get_display (GTK_WIDGET (jp));
+		c = gtk_clipboard_get_for_display (d, GDK_SELECTION_CLIPBOARD);
+		gnome_print_job_preview_get_targets (jp, &t, &nt);
+		gtk_clipboard_set_can_store (c, t, nt);
+		target_entries_free (t, nt);
+		gtk_clipboard_store (c);
+	}
 	gtk_widget_destroy (GTK_WIDGET (jp));
 }
 
@@ -257,12 +277,8 @@ gnome_print_job_preview_clear_undo (GnomePrintJobPreview *jp)
 	gnome_print_job_preview_clear_undo_redo (jp, TRUE);
 }
 
-enum {
-	TARGET_GNOME_PRINT_META
-};
-
 static GtkTargetEntry target_table[] = {
-	{ "GNOME_PRINT_META", 0, TARGET_GNOME_PRINT_META }
+	{ "GNOME_PRINT_META", 0, 0 }
 };
 
 /*
@@ -280,6 +296,9 @@ gnome_print_job_preview_update_pointer (GnomePrintJobPreview *jp, guint n)
 
 	g_return_if_fail (GNOME_IS_PRINT_JOB_PREVIEW (jp));
 
+	if (jp->nx == 0 || jp->ny == 0)
+		return;
+
 	n = MIN (jp->selection->len, n);
 	p = g_array_index (jp->pages, GnomePrintJobPreviewPage, 0);
 	col = (n - p.n) % jp->nx;
@@ -311,23 +330,6 @@ gnome_print_job_preview_update_pointer (GnomePrintJobPreview *jp, guint n)
 	}
 }
 
-static void
-gnome_print_job_preview_pointer_type_changed (GnomePrintJobPreview *jp)
-{
-	g_return_if_fail (GNOME_IS_PRINT_JOB_PREVIEW (jp));
-
-	if (jp->pointer_t &&
-	    !(jp->pointer_l->object.flags & GNOME_CANVAS_ITEM_VISIBLE) &&
-	    !(jp->pointer_r->object.flags & GNOME_CANVAS_ITEM_VISIBLE))
-		gnome_print_job_preview_update_pointer (jp, jp->current_page);
-	else if (!jp->pointer_t &&
-		 ((jp->pointer_l->object.flags & GNOME_CANVAS_ITEM_VISIBLE) ||
-		  (jp->pointer_r->object.flags & GNOME_CANVAS_ITEM_VISIBLE))) {
-		gnome_canvas_item_hide (jp->pointer_l);
-		gnome_canvas_item_hide (jp->pointer_r);
-	}
-}
-
 static void
 gnome_print_job_preview_selection_changed (GnomePrintJobPreview *jp)
 {
@@ -460,6 +462,9 @@ gnome_print_job_preview_show_pages (GnomePrintJobPreview *jp, guint page)
 	guint i;
 
 	g_return_if_fail (GNOME_IS_PRINT_JOB_PREVIEW (jp));
+	if (jp->selection->len == 0)
+		return;
+
 	g_return_if_fail (page < jp->selection->len);
 
 	for (i = 0; i < jp->pages->len; i++)
@@ -909,21 +914,116 @@ gnome_print_job_preview_get_page_at (GnomePrintJobPreview *jp, guint x, guint y)
 	return MIN (row * jp->nx + col, jp->pages->len);
 }
 
+static GdkPixbuf *
+gnome_print_job_preview_get_pixbuf_for_meta (GnomePrintJobPreview *jp,
+		GnomePrintMeta *meta, guint page, gdouble factor)
+{
+	GnomePrintContext *rbuf;
+	GdkPixbuf *pixbuf;
+	gdouble translate[6], scale[6], page2buf[6];
+
+	g_return_val_if_fail (GNOME_IS_PRINT_JOB_PREVIEW (jp), NULL);
+	g_return_val_if_fail (GNOME_IS_PRINT_META (meta), NULL);
+
+	rbuf = gnome_print_context_new_from_module_name ("rbuf");
+	if (!rbuf) return NULL;
+	
+	pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, TRUE, 8,
+			jp->paw * factor, jp->pah * factor);
+	gdk_pixbuf_fill (pixbuf, 0);
+	art_affine_translate (translate, 0, -jp->pah);
+	art_affine_scale (scale, factor, -factor);
+	art_affine_multiply (page2buf, translate, scale);
+	g_object_set (G_OBJECT (rbuf),
+			"pixels",    gdk_pixbuf_get_pixels    (pixbuf),
+			"width",     gdk_pixbuf_get_width     (pixbuf),
+			"height",    gdk_pixbuf_get_height    (pixbuf),
+			"rowstride", gdk_pixbuf_get_rowstride (pixbuf),
+			"alpha",     gdk_pixbuf_get_has_alpha (pixbuf),
+			"page2buf", page2buf,
+			NULL);
+	gnome_print_meta_render_page (meta, rbuf, page, TRUE);
+	g_object_unref (G_OBJECT (rbuf));
+	return pixbuf;
+}
+
+static GdkPixbuf *
+gnome_print_job_preview_get_pixbuf_for_selection (GnomePrintJobPreview *jp,
+		gdouble factor)
+{
+	GnomePrintMeta *meta;
+	guint i;
+
+	g_return_val_if_fail (GNOME_IS_PRINT_JOB_PREVIEW (jp), NULL);
+
+	g_object_get (jp->job, "context", &meta, NULL);
+	for (i = 0; i < jp->selection->len; i++)
+		if (g_array_index (jp->selection, gboolean, i))
+			return gnome_print_job_preview_get_pixbuf_for_meta (jp, meta, i, factor);
+	return NULL;
+}
+
 static void
 on_drag_data_get (GtkWidget *widget, GdkDragContext *context,
 		  GtkSelectionData *selection_data,
 		  guint info, guint time_, GnomePrintJobPreview *jp)
 {
-	GnomePrintContext *meta = gnome_print_meta_new ();
-	guint i;
+	if (selection_data->target == gdk_atom_intern ("GNOME_PRINT_META", FALSE)) {
+		GnomePrintContext *meta = gnome_print_meta_new ();
+		guint i;
 
-	for (i = 0; i < jp->selection->len; i++)
-		if (g_array_index (jp->selection, gboolean, i))
-			gnome_print_job_render_page (jp->job, meta, i, TRUE);
-	gtk_selection_data_set (selection_data, selection_data->target, 8,
-		gnome_print_meta_get_buffer (GNOME_PRINT_META (meta)),
-		gnome_print_meta_get_length (GNOME_PRINT_META (meta)));
-	g_object_unref (G_OBJECT (meta));
+		for (i = 0; i < jp->selection->len; i++)
+			if (g_array_index (jp->selection, gboolean, i))
+				gnome_print_job_render_page (jp->job, meta, i, TRUE);
+		gtk_selection_data_set (selection_data, selection_data->target, 8,
+				gnome_print_meta_get_buffer (GNOME_PRINT_META (meta)),
+				gnome_print_meta_get_length (GNOME_PRINT_META (meta)));
+		g_object_unref (G_OBJECT (meta));
+	} else {
+		GdkPixbuf *p = gnome_print_job_preview_get_pixbuf_for_selection (jp, 1.);
+		gtk_selection_data_set_pixbuf (selection_data, p);
+		g_object_unref (G_OBJECT (p));
+	}
+}
+
+static void
+gnome_print_job_preview_set_state_editing (GnomePrintJobPreview *jp)
+{
+	GnomePrintJobPreviewState s;
+
+	g_return_if_fail (GNOME_IS_PRINT_JOB_PREVIEW (jp));
+
+	if (jp->state == GNOME_PRINT_JOB_PREVIEW_STATE_EDITING) return;
+	s = jp->state; jp->state = GNOME_PRINT_JOB_PREVIEW_STATE_EDITING;
+
+	/* Clean up previous states */
+	if (s == GNOME_PRINT_JOB_PREVIEW_STATE_DRAGGING) {
+		GdkDisplay *d = gtk_widget_get_display (GTK_WIDGET (jp));
+
+		gdk_display_pointer_ungrab (d, jp->event->button.time);
+		gnome_print_job_preview_select_page (jp,
+			gnome_print_job_preview_get_page_at (jp,
+				jp->event->button.x, jp->event->button.y));
+	} else
+		gnome_print_job_preview_select_page (jp, jp->current_page);
+
+	/* Enable drag & drop */
+	gtk_drag_source_set (GTK_WIDGET (jp->canvas),
+		GDK_BUTTON1_MASK | GDK_BUTTON3_MASK, target_table,
+		G_N_ELEMENTS (target_table),
+		GDK_ACTION_COPY | GDK_ACTION_MOVE);
+	gtk_drag_source_add_image_targets (GTK_WIDGET (jp->canvas));
+	if (s == GNOME_PRINT_JOB_PREVIEW_STATE_DRAGGING) {
+		gtk_drag_begin (GTK_WIDGET (jp->canvas),
+			gtk_drag_source_get_target_list (GTK_WIDGET (jp->canvas)),
+			(jp->event->button.state & GDK_SHIFT_MASK) ?
+				GDK_ACTION_MOVE : GDK_ACTION_COPY,
+			jp->event->button.button, jp->event);
+	}
+
+	if (!gtk_toggle_action_get_active (jp->other.edit))
+		gtk_toggle_action_set_active (jp->other.edit, TRUE);
+	gtk_widget_grab_focus (GTK_WIDGET (jp->canvas));
 }
 
 static void
@@ -934,7 +1034,13 @@ gnome_print_job_preview_set_pointer_type (GnomePrintJobPreview *jp,
 
 	if (jp->pointer_t & t) return;
 	jp->pointer_t |= t;
-	gnome_print_job_preview_pointer_type_changed (jp);
+	if (!jp->pointer_t) return;
+	if ((jp->pointer_l->object.flags & GNOME_CANVAS_ITEM_VISIBLE) ||
+			(jp->pointer_r->object.flags & GNOME_CANVAS_ITEM_VISIBLE)) return;
+	
+	gnome_print_job_preview_set_state_editing (jp);
+	g_object_set (G_OBJECT (jp->ccp.paste), "sensitive", TRUE, NULL);
+	gnome_print_job_preview_update_pointer (jp, jp->current_page);
 }
 
 static void
@@ -946,15 +1052,27 @@ gnome_print_job_preview_unset_pointer_type (GnomePrintJobPreview *jp,
 	if (!(jp->pointer_t & t)) return;
 
 	jp->pointer_t &= ~t;
-	gnome_print_job_preview_pointer_type_changed (jp);
+	if (jp->pointer_t) return;
+	if (jp->pointer_l->object.flags & GNOME_CANVAS_ITEM_VISIBLE)
+		gnome_canvas_item_hide (jp->pointer_l);
+	if (jp->pointer_r->object.flags & GNOME_CANVAS_ITEM_VISIBLE)
+		gnome_canvas_item_hide (jp->pointer_r);
+	g_object_set (G_OBJECT (jp->ccp.paste), "sensitive", FALSE, NULL);
 }
 
 static void
 on_drag_begin (GtkWidget *widget, GdkDragContext *context,
 	       GnomePrintJobPreview *jp)
 {
+	GdkPixbuf *pixbuf;
+
 	gnome_print_job_preview_set_pointer_type (jp,
 			GNOME_PRINT_JOB_PREVIEW_POINTER_TYPE_DRAG_DROP);
+
+	/* Show a preview if possible */
+	pixbuf = gnome_print_job_preview_get_pixbuf_for_selection (jp, 0.1);
+	gtk_drag_set_icon_pixbuf (context, pixbuf, 0, 0);
+	g_object_unref (G_OBJECT (pixbuf));
 }
 
 static gboolean
@@ -1006,8 +1124,9 @@ gnome_print_job_preview_set_state_normal (GnomePrintJobPreview *jp)
 	/* Clean up previous states. */
 	gnome_print_job_preview_select_all_none (jp, FALSE);
 	if (jp->pointer_t) {
-		jp->pointer_t = GNOME_PRINT_JOB_PREVIEW_POINTER_TYPE_NONE;
-		gnome_print_job_preview_pointer_type_changed (jp);
+		gnome_print_job_preview_unset_pointer_type (jp,
+				GNOME_PRINT_JOB_PREVIEW_POINTER_TYPE_CUT_COPY |
+				GNOME_PRINT_JOB_PREVIEW_POINTER_TYPE_DRAG_DROP);
 	}
 	if (s == GNOME_PRINT_JOB_PREVIEW_STATE_EDITING)
 		gtk_drag_source_unset (GTK_WIDGET (jp->canvas));
@@ -1049,56 +1168,18 @@ gnome_print_job_preview_set_state_dragging (GnomePrintJobPreview *jp)
 	gdk_cursor_unref (c);
 }
 
-static void
-gnome_print_job_preview_set_state_editing (GnomePrintJobPreview *jp)
-{
-	GnomePrintJobPreviewState s;
-
-	g_return_if_fail (GNOME_IS_PRINT_JOB_PREVIEW (jp));
-
-	if (jp->state == GNOME_PRINT_JOB_PREVIEW_STATE_EDITING) return;
-	s = jp->state; jp->state = GNOME_PRINT_JOB_PREVIEW_STATE_EDITING;
-
-	/* Clean up previous states */
-	if (s == GNOME_PRINT_JOB_PREVIEW_STATE_DRAGGING) {
-		GdkDisplay *d = gtk_widget_get_display (GTK_WIDGET (jp));
-
-		gdk_display_pointer_ungrab (d, jp->event->button.time);
-		gnome_print_job_preview_select_page (jp,
-			gnome_print_job_preview_get_page_at (jp,
-				jp->event->button.x, jp->event->button.y));
-	} else
-		gnome_print_job_preview_select_page (jp, jp->current_page);
-
-	/* Enable drag & drop */
-	gtk_drag_source_set (GTK_WIDGET (jp->canvas),
-		GDK_BUTTON1_MASK | GDK_BUTTON3_MASK, target_table,
-		G_N_ELEMENTS (target_table),
-		GDK_ACTION_COPY | GDK_ACTION_MOVE);
-	if (s == GNOME_PRINT_JOB_PREVIEW_STATE_DRAGGING) {
-		gtk_drag_begin (GTK_WIDGET (jp->canvas),
-			gtk_drag_source_get_target_list (GTK_WIDGET (jp->canvas)),
-			GDK_ACTION_MOVE, jp->event->button.button, jp->event);
-		gdk_event_free (jp->event);
-		jp->event = NULL;
-	}
-
-	if (!gtk_toggle_action_get_active (jp->other.edit))
-		gtk_toggle_action_set_active (jp->other.edit, TRUE);
-	gtk_widget_grab_focus (GTK_WIDGET (jp->canvas));
-}
-
 static gint
 on_canvas_button_press_event (GtkWidget *widget, GdkEventButton *event,
 			      GnomePrintJobPreview *jp)
 {
 	guint n;
-	
+
+	if (jp->event) gdk_event_free (jp->event);
+	jp->event = gdk_event_copy ((GdkEvent *) event);
+
 	if (event->button == 1) {
 		switch (jp->state) {
 		case GNOME_PRINT_JOB_PREVIEW_STATE_NORMAL:
-			if (jp->event) gdk_event_free (jp->event);
-			jp->event = gdk_event_copy ((GdkEvent *) event);
 			gnome_print_job_preview_set_state_dragging (jp);
 			return TRUE;
 		case GNOME_PRINT_JOB_PREVIEW_STATE_EDITING:
@@ -1133,7 +1214,7 @@ on_canvas_button_press_event (GtkWidget *widget, GdkEventButton *event,
 			return FALSE;
 		}
 	}
-		
+
 	return FALSE;
 }
 
@@ -1162,10 +1243,6 @@ on_canvas_motion_notify_event (GtkWidget *widget, GdkEventMotion *event,
 		dx = jp->anchorx - x;
 		dy = jp->anchory - y;
 
-		/*
-		 * Start dragging? Only do so if the user clearly goes beyond
-		 * the limits of the canvas.
-		 */
 		if (!dx && !dy) return TRUE;
 		width  = jp->canvas->layout.width;
 		height = jp->canvas->layout.height;
@@ -1175,16 +1252,24 @@ on_canvas_motion_notify_event (GtkWidget *widget, GdkEventMotion *event,
 		bottom = MIN (height, top  + GTK_WIDGET (jp->canvas)->allocation.height);
 		if ((dx < -left) || (right  + dx > width) ||
 		    (dy < -top ) || (bottom + dy > height)) {
-			if ((left   + dx < -0.1 * width ) ||
-			    (top    + dy < -0.1 * height) ||
-			    (right  + dx >  1.1 * width ) ||
-			    (bottom + dy >  1.1 * height)) {
+			guint p_old, p_new;
+			GtkAllocation a = GTK_WIDGET (jp->canvas)->allocation;
+
+			p_old = gnome_print_job_preview_get_page_at (jp, jp->anchorx, jp->anchory);
+			p_new = gnome_print_job_preview_get_page_at (jp, x, y);
+
+			/*
+			 * Start dragging? Only if the user moves over another page or 
+			 * out of the canvas.
+			 */
+			if (((p_old != p_new) && (p_old + 1 != p_new)) ||
+					(x < 0) || (x > a.width) ||
+					(y < 0) || (y > a.height)) {
 				gnome_print_job_preview_set_state_editing (jp);
 				return FALSE;
-			} else {
-				dx = MIN (width  - right , MAX (dx, -left));
-				dy = MIN (height - bottom, MAX (dy, -top ));
 			}
+			dx = MIN (width  - right , MAX (dx, -left));
+			dy = MIN (height - bottom, MAX (dy, -top ));
 		}
 		if (!dx && !dy) return TRUE;
 
@@ -1241,25 +1326,33 @@ gnome_print_job_preview_goto_previous_screen (GnomePrintJobPreview *jp)
 }
 
 static void
-clipboard_get_func (GtkClipboard *c, GtkSelectionData *selection_data,
+clipboard_get_func (GtkClipboard *c, GtkSelectionData *sd,
 		    guint info, GnomePrintJobPreview *jp)
 {
-	gtk_selection_data_set (selection_data, TARGET_GNOME_PRINT_META, 8,
-		gnome_print_meta_get_buffer (GNOME_PRINT_META (jp->clipboard)),
-		gnome_print_meta_get_length (GNOME_PRINT_META (jp->clipboard)));
+	if (sd->target == gdk_atom_intern ("GNOME_PRINT_META", FALSE)) {
+		gtk_selection_data_set (sd, sd->target, 8,
+				gnome_print_meta_get_buffer (GNOME_PRINT_META (jp->clipboard)),
+				gnome_print_meta_get_length (GNOME_PRINT_META (jp->clipboard)));
+	} else {
+		GdkPixbuf *p = gnome_print_job_preview_get_pixbuf_for_meta (jp, 
+				GNOME_PRINT_META (jp->clipboard), 0, 1.);
+		gtk_selection_data_set_pixbuf (sd, p);
+		g_object_unref (G_OBJECT (p));
+	}
 }
 
 static void
-clipboard_received_func (GtkClipboard *c, GtkSelectionData *selection_data,
+clipboard_received_func (GtkClipboard *c, GtkSelectionData *sd,
 			 GnomePrintJobPreview *jp)
 {
-	GnomePrintContext *meta = gnome_print_meta_new ();
+	if (sd->target == gdk_atom_intern ("GNOME_PRINT_META", FALSE)) {
+		GnomePrintContext *meta = gnome_print_meta_new (); 
 
-	gnome_print_meta_render_data (meta, selection_data->data,
-				      selection_data->length);
-	gnome_print_job_preview_cmd_insert (jp, jp->current_page,
-					    GNOME_PRINT_META (meta));
-	g_object_unref (G_OBJECT (meta));
+		gnome_print_meta_render_data (meta, sd->data, sd->length);
+		gnome_print_job_preview_cmd_insert (jp, jp->current_page,
+				GNOME_PRINT_META (meta));
+		g_object_unref (G_OBJECT (meta));
+	}
 }
 
 static void
@@ -1276,24 +1369,59 @@ gnome_print_job_preview_paste (GnomePrintJobPreview *jp)
 {
 	GdkDisplay *d = gtk_widget_get_display (GTK_WIDGET (jp));
 	GtkClipboard *c = gtk_clipboard_get_for_display (d, GDK_SELECTION_CLIPBOARD);
-	GdkAtom atom = gdk_atom_intern ("GNOME_PRINT_META", FALSE);
 
 	g_return_if_fail (GNOME_IS_PRINT_JOB_PREVIEW (jp));
 
-	gtk_clipboard_request_contents (c, atom, 
-		(GtkClipboardReceivedFunc) clipboard_received_func, jp);
+	gtk_clipboard_request_contents (c, 
+			gdk_atom_intern ("GNOME_PRINT_META", FALSE),
+			(GtkClipboardReceivedFunc) clipboard_received_func, jp);
+}
+
+static void
+target_entries_free (GtkTargetEntry *t, guint nt)
+{
+	guint i;
+
+	for (i = 0; i < nt; i++) g_free (t[i].target);
+	g_free (t);
+}
+
+static void
+gnome_print_job_preview_get_targets (GnomePrintJobPreview *jp,
+		GtkTargetEntry **t, guint *nt)
+{
+	GtkTargetList *list = NULL;
+	guint i;
+
+	g_return_if_fail (GNOME_IS_PRINT_JOB_PREVIEW (jp));
+	g_return_if_fail (t);
+	g_return_if_fail (nt);
+
+	list = gtk_target_list_new (target_table, G_N_ELEMENTS (target_table));
+	gtk_target_list_add_image_targets (list, 0, FALSE);
+	*nt = g_list_length (list->list);
+	*t = g_new (GtkTargetEntry, *nt);
+	for (i = 0; i < *nt; i++) {
+		GtkTargetPair *pair = g_list_nth_data (list->list, i);
+		(*t)[i].target = gdk_atom_name (pair->target);
+		(*t)[i].flags = pair->flags;
+		(*t)[i].info = pair->info;
+	}
+	gtk_target_list_unref (list);
 }
 
 static void
 gnome_print_job_preview_cut_copy (GnomePrintJobPreview *jp, gboolean cut)
 {
-	GdkDisplay *d = gtk_widget_get_display (GTK_WIDGET (jp));
-	GtkClipboard *c = gtk_clipboard_get_for_display (d, GDK_SELECTION_CLIPBOARD);
+	GdkDisplay *d;
+	GtkClipboard *c;
 	guint i;
 	GnomePrintMeta *meta;
+	GtkTargetEntry *t = NULL;
+	guint nt = 0;
 
 	g_return_if_fail (GNOME_IS_PRINT_JOB_PREVIEW (jp));
-	
+
 	if (!g_array_index (jp->selection, gboolean,
 			MIN (jp->current_page, jp->selection->len - 1)))
 		gnome_print_job_preview_select_page (jp, jp->current_page);
@@ -1305,11 +1433,14 @@ gnome_print_job_preview_cut_copy (GnomePrintJobPreview *jp, gboolean cut)
 		if (g_array_index (jp->selection, gboolean, i))
 			gnome_print_meta_render_page (meta,
 						jp->clipboard, i, TRUE);
-	gtk_clipboard_set_with_owner (c, target_table,
-			G_N_ELEMENTS (target_table),
+
+	gnome_print_job_preview_get_targets (jp, &t, &nt);
+	d = gtk_widget_get_display (GTK_WIDGET (jp));
+	c = gtk_clipboard_get_for_display (d, GDK_SELECTION_CLIPBOARD);
+	gtk_clipboard_set_with_owner (c, t, nt, 
 			(GtkClipboardGetFunc  ) clipboard_get_func,
-			(GtkClipboardClearFunc) clipboard_clear_func,
-			G_OBJECT (jp));
+			(GtkClipboardClearFunc) clipboard_clear_func, G_OBJECT (jp));
+	target_entries_free (t, nt);
 	if (cut) gnome_print_job_preview_cmd_delete (jp);
 
 	gnome_print_job_preview_set_pointer_type (jp,
@@ -1434,8 +1565,11 @@ gnome_print_job_preview_nx_and_ny_changed (GnomePrintJobPreview *jp)
 		GnomePrintJobPreviewPage p;
 		
 		p = g_array_index (jp->pages, GnomePrintJobPreviewPage, i);
-		col = i % jp->nx;
-		row = i / jp->nx;
+		if (jp->nx > 0) {
+			col = i % jp->nx;
+			row = i / jp->nx;
+		} else
+			col = row = 0;
 		g_object_set (p.group,
 			"x", (gdouble) col * (jp->paw + PAGE_PAD * 2),
 			"y", (gdouble) row * (jp->pah + PAGE_PAD * 2), NULL);
@@ -1754,6 +1888,27 @@ on_canvas_key_press_event (GtkWidget *widget, GdkEventKey *event,
 			return TRUE;
 		}
 		break;
+	case GDK_c:
+		if (event->state & GDK_CONTROL_MASK) {
+			gnome_print_job_preview_set_state_editing (jp);
+			gnome_print_job_preview_cut_copy (jp, FALSE);
+			return TRUE;
+		}
+		break;
+	case GDK_x:
+		if (event->state & GDK_CONTROL_MASK) {
+			gnome_print_job_preview_set_state_editing (jp);
+			gnome_print_job_preview_cut_copy (jp, TRUE);
+			return TRUE;
+		}
+		break;
+	case GDK_v:
+		if (event->state & GDK_CONTROL_MASK) {
+			gnome_print_job_preview_set_state_editing (jp);
+			gnome_print_job_preview_paste (jp);
+			return TRUE;
+		}
+		break;
 	case GDK_space:
 		if (jp->state == GNOME_PRINT_JOB_PREVIEW_STATE_EDITING) {
 			if (gnome_print_job_preview_count_selected (jp) > 1) {
@@ -1850,19 +2005,31 @@ on_drag_data_received (GtkWidget *widget, GdkDragContext *context,
 		       gint x, gint y, GtkSelectionData *data,
 		       guint info, guint time_, GnomePrintJobPreview *jp)
 {
-	guint n = gnome_print_job_preview_get_page_at (jp, x, y);
+	guint n, no;
+	GtkWidget *w, *wo;
+	GnomePrintContext *meta;
 
-	if ((context->action & GDK_ACTION_MOVE) && 
-	    gtk_drag_get_source_widget (context) == widget) {
-		gnome_print_job_preview_cmd_move (jp, n);
-	} else {
-		GnomePrintContext *meta = gnome_print_meta_new ();
+	n = gnome_print_job_preview_get_page_at (jp, x, y);
+	no = gnome_print_job_preview_get_page_at (jp, jp->event->button.x,
+			jp->event->button.y);
 
-		gnome_print_meta_render_data (meta, data->data, data->length);
-		gnome_print_job_preview_cmd_insert (jp, n,
-						    GNOME_PRINT_META (meta));
-		g_object_unref (G_OBJECT (meta));
+	w = widget;
+	wo = gtk_drag_get_source_widget (context);
+
+	if ((w == wo) && (n == no)) {
+		gtk_drag_finish (context, FALSE, FALSE, time_);
+		return;
 	}
+
+	if ((w == wo) && (context->action & GDK_ACTION_MOVE)) {
+		gnome_print_job_preview_cmd_move (jp, n);
+		return;
+	}
+
+	meta = gnome_print_meta_new ();
+	gnome_print_meta_render_data (meta, data->data, data->length);
+	gnome_print_job_preview_cmd_insert (jp, n, GNOME_PRINT_META (meta));
+	g_object_unref (G_OBJECT (meta));
 }
 
 static gboolean
@@ -1982,7 +2149,7 @@ gnome_print_job_preview_parse_layout (GnomePrintJobPreview *jp)
 {
 	GnomePrintConfig *config;
 	GnomePrintLayoutData *lyd;
-	guint w = GPMP_A4_WIDTH, h = GPMP_A4_HEIGHT;
+	guint w = (210.0 * 72.0 / 25.4), h = (297.0 * 72.0 / 2.54);
 
 	/* Calculate layout-compensated page dimensions */
 	art_affine_identity (jp->pa2ly);
@@ -2070,6 +2237,11 @@ gnome_print_job_preview_destroy (GtkObject *object)
 {
 	GnomePrintJobPreview *jp = GNOME_PRINT_JOB_PREVIEW (object);
 
+	if (jp->clipboard) {
+		g_object_unref (G_OBJECT (jp->clipboard));
+		jp->clipboard = NULL;
+	}
+
 	if (jp->undo) {
 		gnome_print_job_preview_clear_undo (jp);
 		g_array_free (jp->undo, TRUE);
@@ -2083,11 +2255,6 @@ gnome_print_job_preview_destroy (GtkObject *object)
 
 	gnome_print_job_preview_set_job (jp, NULL);
 
-	if (jp->clipboard) {
-		g_object_unref (G_OBJECT (jp->clipboard));
-		jp->clipboard = NULL;
-	}
-
 	if (jp->pages != NULL) {
 		g_array_free (jp->pages, TRUE);
 		jp->pages = NULL;
@@ -2111,12 +2278,6 @@ gnome_print_job_preview_destroy (GtkObject *object)
 	GTK_OBJECT_CLASS (parent_class)->destroy (object);
 }
 
-static void
-gnome_print_job_preview_finalize (GObject *object)
-{
-	G_OBJECT_CLASS (parent_class)->finalize (object);
-}
-
 enum {
 	PROP_0,
 	PROP_NX,
@@ -2190,7 +2351,6 @@ gnome_print_job_preview_class_init (GnomePrintJobPreviewClass *klass)
 
 	object_class->destroy = gnome_print_job_preview_destroy;
 
-	gobject_class->finalize     = gnome_print_job_preview_finalize;
 	gobject_class->set_property = gnome_print_job_preview_set_property;
 	gobject_class->get_property = gnome_print_job_preview_get_property;
 
@@ -2214,34 +2374,57 @@ on_realize (GtkWidget *widget, GnomePrintJobPreview *jp)
 	gnome_print_job_preview_zoom (jp, -1.);
 }
 
-#define GNOME_PRINT_JOB_PREVIEW_UI_DEFINITION			\
-"<ui>"								\
-"  <toolbar name=\"StandardToolbar\">"				\
-"    <toolitem action=\"close\"/>"				\
-"    <toolitem action=\"print\"/>"				\
-"    <separator/>"						\
-"    <toolitem action=\"cut\"/>"				\
-"    <toolitem action=\"copy\"/>"				\
-"    <toolitem action=\"paste\"/>"				\
-"    <separator/>"						\
-"    <toolitem action=\"undo\"/>"				\
-"    <toolitem action=\"redo\"/>"				\
-"    <separator/>"						\
-"    <toolitem action=\"first\"/>"				\
-"    <toolitem action=\"previous\"/>"				\
-"    <toolitem action=\"next\"/>"				\
-"    <toolitem action=\"last\"/>"				\
-"    <separator/>"						\
-"    <toolitem action=\"zoom_100\"/>"				\
-"    <toolitem action=\"zoom_fit\"/>"				\
-"    <toolitem action=\"zoom_in\"/>"				\
-"    <toolitem action=\"zoom_out\"/>"				\
-"    <separator/>"						\
-"    <toolitem action=\"multi\"/>"				\
-"    <toolitem action=\"edit\"/>"				\
-"    <toolitem action=\"theme\"/>"				\
-"  </toolbar>"							\
-"</ui>"
+static void
+cb_clipboard_targets_changed (GtkClipboard *clipboard, GdkAtom *targets,
+		gint n_targets, gpointer data)
+{
+	GnomePrintJobPreview *jp = GNOME_PRINT_JOB_PREVIEW (data);
+	guint i;
+	
+	for (i = 0; (gint) i < n_targets; i++)
+		if (targets[i] == gdk_atom_intern ("GNOME_PRINT_META", TRUE)) {
+			gnome_print_job_preview_set_pointer_type (jp,
+					GNOME_PRINT_JOB_PREVIEW_POINTER_TYPE_CUT_COPY);
+			return;
+		}
+	gnome_print_job_preview_unset_pointer_type (jp,
+			GNOME_PRINT_JOB_PREVIEW_POINTER_TYPE_CUT_COPY);
+}
+
+static void
+gnome_print_job_preview_check_clipboard (GnomePrintJobPreview *jp)
+{
+	GdkDisplay *d;
+	GtkClipboard *c;
+
+	g_return_if_fail (GNOME_IS_PRINT_JOB_PREVIEW (jp));
+
+	if (jp->state != GNOME_PRINT_JOB_PREVIEW_STATE_EDITING) return;
+	d = gtk_widget_get_display (GTK_WIDGET (jp));
+	c = gtk_clipboard_get_for_display (d, GDK_SELECTION_CLIPBOARD);
+	gtk_clipboard_request_targets (c, cb_clipboard_targets_changed, jp);
+}
+
+static gboolean
+on_focus_in_event (GtkWidget *widget, GdkEventFocus *event)
+{
+	GnomePrintJobPreview *jp = GNOME_PRINT_JOB_PREVIEW (widget);
+
+	gnome_print_job_preview_check_clipboard (jp);
+
+	return FALSE;
+}
+
+static gboolean
+on_focus_out_event (GtkWidget *widget, GdkEventFocus *event)
+{
+	GnomePrintJobPreview *jp = GNOME_PRINT_JOB_PREVIEW (widget);
+
+	/* Hide the pointer */
+	gnome_print_job_preview_unset_pointer_type (jp,
+			GNOME_PRINT_JOB_PREVIEW_POINTER_TYPE_CUT_COPY);
+	return FALSE;
+}
 
 static void
 gnome_print_job_preview_init (GnomePrintJobPreview *jp)
@@ -2286,6 +2469,9 @@ gnome_print_job_preview_init (GnomePrintJobPreview *jp)
 	jp->ccp.cut = gtk_action_new ("cut", _("Cut"), _("Cut"), GTK_STOCK_CUT);
 	jp->ccp.copy = gtk_action_new ("copy", _("Copy"), _("Copy"), GTK_STOCK_COPY);
 	jp->ccp.paste = gtk_action_new ("paste", _("Paste"),  _("Paste"), GTK_STOCK_PASTE);
+	g_object_set (G_OBJECT (jp->ccp.cut), "sensitive", FALSE, NULL);
+	g_object_set (G_OBJECT (jp->ccp.copy), "sensitive", FALSE, NULL);
+	g_object_set (G_OBJECT (jp->ccp.paste), "sensitive", FALSE, NULL);
 	gtk_action_group_add_action_with_accel (jp->ccp.group, jp->ccp.cut, NULL);
 	gtk_action_group_add_action_with_accel (jp->ccp.group, jp->ccp.copy, NULL);
 	gtk_action_group_add_action_with_accel (jp->ccp.group, jp->ccp.paste, NULL);
@@ -2361,13 +2547,24 @@ gnome_print_job_preview_init (GnomePrintJobPreview *jp)
 		G_N_ELEMENTS (target_table),
 		GDK_ACTION_COPY | GDK_ACTION_MOVE);
 
-	gtk_ui_manager_add_ui_from_string (jp->ui_manager,
-		GNOME_PRINT_JOB_PREVIEW_UI_DEFINITION, -1, NULL);
+	/* Toolbar */
+	gtk_ui_manager_add_ui_from_file (jp->ui_manager,
+			GNOME_PRINT_JOB_PREVIEW_DATADIR "/gnome-print-job-preview.xml", &e);
 	if (!e)
 		tb = gtk_ui_manager_get_widget (jp->ui_manager, "/StandardToolbar");
 	if (e || !tb) {
-		tb = gtk_label_new (e->message);
-		g_error_free (e);
+		gchar *txt;
+
+		/* Should that error message be translated? */
+		txt = g_strdup_printf ("The toolbar can not be displayed: %s",
+				e ? e->message : "Path '/StandardToolbar' not found");
+		tb = gtk_label_new (txt);
+		g_free (txt);
+		gtk_label_set_line_wrap (GTK_LABEL (tb), TRUE);
+		if (e) {
+			g_error_free (e);
+			e = NULL;
+		}
 	} else {
 		gtk_toolbar_set_style (GTK_TOOLBAR (tb), GTK_TOOLBAR_ICONS);
 		gtk_toolbar_set_orientation (GTK_TOOLBAR (tb), GTK_ORIENTATION_HORIZONTAL);
@@ -2436,6 +2633,8 @@ gnome_print_job_preview_init (GnomePrintJobPreview *jp)
 	g_signal_connect (G_OBJECT (jp->page_entry), "insert_text", G_CALLBACK (entry_insert_text_cb), jp);
 	g_signal_connect (G_OBJECT (jp->page_entry), "focus_out_event", G_CALLBACK (entry_focus_out_event_cb), jp);
 	g_signal_connect (G_OBJECT (jp), "delete_event", G_CALLBACK (on_delete_event), jp);
+	g_signal_connect (G_OBJECT (jp), "focus_in_event", G_CALLBACK (on_focus_in_event), jp);
+	g_signal_connect (G_OBJECT (jp), "focus_out_event", G_CALLBACK (on_focus_out_event), jp);
 }
 
 GType
@@ -2458,46 +2657,12 @@ gnome_print_job_preview_get_type (void)
 	return type;
 }
 
-static void
-allow_cut_copy (GnomePrintJobPreview *jp, gboolean allow)
-{
-	g_object_set (G_OBJECT (jp->ccp.paste),
-		"sensitive", allow_cut_copy,
-		NULL);
-	if (allow)
-		gnome_print_job_preview_set_pointer_type (jp,
-			GNOME_PRINT_JOB_PREVIEW_POINTER_TYPE_CUT_COPY);
-	else
-		gnome_print_job_preview_unset_pointer_type (jp,
-			GNOME_PRINT_JOB_PREVIEW_POINTER_TYPE_CUT_COPY);
-
-}
-
-static void
-cb_clipboard_targets_changed (GtkClipboard     *clipboard,
-			      GdkAtom          *targets,
-			      gint              n_targets,
-			      gpointer		jp)
-{
-	guint i;
-	for (i = 0; i < n_targets; i++)
-		if (targets[i] == gdk_atom_intern ("GNOME_PRINT_META", TRUE)) {
-			allow_cut_copy (GNOME_PRINT_JOB_PREVIEW (jp), TRUE);
-			return;
-		}
-	allow_cut_copy (jp, FALSE);
-}
-
 static gboolean
 cb_clipboard_owner_changed (GtkClipboard *clipboard,
 			    G_GNUC_UNUSED GdkEventOwnerChange *event,
 			    GnomePrintJobPreview *jp)
 {
-	if (jp->state == GNOME_PRINT_JOB_PREVIEW_STATE_EDITING)
-		gtk_clipboard_request_targets (clipboard,
-			cb_clipboard_targets_changed, jp);
-	else
-		allow_cut_copy (jp, FALSE);
+	gnome_print_job_preview_check_clipboard (jp);
 	return TRUE;
 }
 
