@@ -37,6 +37,7 @@
 #include <libgnomeprint/private/gnome-print-config-private.h>
 
 #include "gnome-print-i18n.h"
+#include "gnome-print-dialog.h"
 #include "gnome-print-paper-selector.h"
 #include "gnome-print-unit-selector.h"
 
@@ -391,7 +392,7 @@ gnome_paper_selector_printer_changed_cb (GPANode *node, guint flags, GnomePaperS
 static void
 gnome_paper_selector_construct (GnomePaperSelector *ps)
 {
-	GtkWidget *vb, *f, *t, *l, *s;
+	GtkWidget *vb, *f, *t, *l, *s, *label;
 	GtkWidget *margin_table, *margin_label;
 	gdouble ml, mr, mt, mb, pw = 1., ph = 1.;
 	AtkObject *atko;
@@ -439,12 +440,12 @@ gnome_paper_selector_construct (GnomePaperSelector *ps)
 	/* Create frame for selection menus */
 	f = gtk_frame_new ("");
 	gtk_frame_set_shadow_type (GTK_FRAME (f), GTK_SHADOW_NONE);
-	l = gtk_label_new ("");
+	label = gtk_label_new ("");
 	text = g_strdup_printf ("<b>%s</b>", _("Paper and Layout"));
-	gtk_label_set_markup (GTK_LABEL (l), text);
+	gtk_label_set_markup (GTK_LABEL (label), text);
 	g_free (text);
-	gtk_frame_set_label_widget (GTK_FRAME (f), l);
-	gtk_widget_show (l);
+	gtk_frame_set_label_widget (GTK_FRAME (f), label);
+	gtk_widget_show (label);
 	gtk_widget_show (f);
 	gtk_box_pack_start (GTK_BOX (vb), f, FALSE, FALSE, 0);
 
@@ -465,6 +466,9 @@ gnome_paper_selector_construct (GnomePaperSelector *ps)
 	ps->pmenu = gpa_option_menu_new (ps->config, GNOME_PRINT_KEY_PAPER_SIZE);
 	gtk_widget_show (ps->pmenu);
 	gtk_table_attach_defaults (GTK_TABLE (t), ps->pmenu, 1, 4, 0, 1);
+	gnome_print_set_atk_relation (label, GTK_WIDGET (GPA_OPTION_MENU (ps->pmenu)->menu));
+	gnome_print_set_atk_relation (l, GTK_WIDGET (GPA_OPTION_MENU (ps->pmenu)->menu));
+
 	gtk_label_set_mnemonic_widget ((GtkLabel *) l, GPA_OPTION_MENU (ps->pmenu)->menu);
 
 	/* Custom paper Width */
@@ -479,6 +483,8 @@ gnome_paper_selector_construct (GnomePaperSelector *ps)
 				     TRUE);
 	gtk_widget_show (s);
 	gtk_table_attach_defaults (GTK_TABLE (t), s, 2, 3, 1, 2);
+	gnome_print_set_atk_relation (label, GTK_WIDGET (ps->p.w->spinbutton));
+	gnome_print_set_atk_relation (l, GTK_WIDGET (ps->p.w->spinbutton));
 	gtk_label_set_mnemonic_widget ((GtkLabel *) l, ps->p.w->spinbutton);
 
 	/* Custom paper Height */
@@ -493,12 +499,17 @@ gnome_paper_selector_construct (GnomePaperSelector *ps)
 				     TRUE);
 	gtk_widget_show (s);
 	gtk_table_attach_defaults (GTK_TABLE (t), s, 2, 3, 2, 3);
+	gnome_print_set_atk_relation (label, GTK_WIDGET (ps->p.h->spinbutton));
+	gnome_print_set_atk_relation (l, GTK_WIDGET (ps->p.h->spinbutton));
 	gtk_label_set_mnemonic_widget ((GtkLabel *) l, ps->p.h->spinbutton);
 
 	/* Custom paper Unit selector */
 	s = gnome_print_unit_selector_new (GNOME_PRINT_UNIT_ABSOLUTE);
 	ps->us = GNOME_PRINT_UNIT_SELECTOR (s);
 	gtk_table_attach_defaults (GTK_TABLE (t), s, 3, 4, 1, 2);
+	l = g_object_get_data (G_OBJECT(s), "combo");
+	if (l !=NULL)
+		gnome_print_set_atk_relation (label, GTK_WIDGET (l));
 	atko = gtk_widget_get_accessible (s);
 	atk_object_set_name (atko, _("Metric selector"));
 	atk_object_set_description (atko, _("Specifies the metric to use when setting the width and height of the paper"));
@@ -510,6 +521,8 @@ gnome_paper_selector_construct (GnomePaperSelector *ps)
 
 	ps->pomenu = gpa_option_menu_new (ps->config, GNOME_PRINT_KEY_PAPER_ORIENTATION);
 	gtk_table_attach_defaults (GTK_TABLE (t), ps->pomenu, 1, 4, 3, 4);
+	gnome_print_set_atk_relation (label, GTK_WIDGET (GPA_OPTION_MENU (ps->pomenu)->menu));
+	gnome_print_set_atk_relation (l, GTK_WIDGET (GPA_OPTION_MENU (ps->pomenu)->menu));
 	gtk_label_set_mnemonic_widget ((GtkLabel *) l, GPA_OPTION_MENU (ps->pomenu)->menu);
 
 	if (ps->flags & GNOME_PAPER_SELECTOR_FEED_ORIENTATION || TRUE) {
@@ -526,6 +539,8 @@ gnome_paper_selector_construct (GnomePaperSelector *ps)
 	ps->lomenu = gpa_option_menu_new (ps->config, GNOME_PRINT_KEY_PAGE_ORIENTATION);
 	gtk_widget_show (ps->lomenu);
 	gtk_table_attach_defaults (GTK_TABLE (t), ps->lomenu, 1, 4, 4, 5);
+	gnome_print_set_atk_relation (label, GTK_WIDGET (GPA_OPTION_MENU (ps->lomenu)->menu));
+	gnome_print_set_atk_relation (l, GTK_WIDGET (GPA_OPTION_MENU (ps->lomenu)->menu));
 	gtk_label_set_mnemonic_widget ((GtkLabel *) l, GPA_OPTION_MENU (ps->lomenu)->menu);
 
 	/* Layout */
@@ -537,6 +552,8 @@ gnome_paper_selector_construct (GnomePaperSelector *ps)
 	ps->lymenu = gpa_option_menu_new (ps->config, GNOME_PRINT_KEY_LAYOUT);
 	gtk_widget_show (ps->lymenu);
 	gtk_table_attach_defaults (GTK_TABLE (t), ps->lymenu, 1, 4, 5, 6);
+	gnome_print_set_atk_relation (label, GTK_WIDGET (GPA_OPTION_MENU (ps->lymenu)->menu));
+	gnome_print_set_atk_relation (l, GTK_WIDGET (GPA_OPTION_MENU (ps->lymenu)->menu));
 	gtk_label_set_mnemonic_widget ((GtkLabel *) l, GPA_OPTION_MENU (ps->lymenu)->menu);
 
 	/* Paper source */
@@ -548,6 +565,8 @@ gnome_paper_selector_construct (GnomePaperSelector *ps)
 	ps->trmenu = gpa_option_menu_new (ps->config, GNOME_PRINT_KEY_PAPER_SOURCE);
 	gtk_widget_show(ps->trmenu);
 	gtk_table_attach_defaults (GTK_TABLE (t), ps->trmenu, 1, 4, 6, 7);
+	gnome_print_set_atk_relation (label, GTK_WIDGET (GPA_OPTION_MENU (ps->trmenu)->menu));
+	gnome_print_set_atk_relation (l, GTK_WIDGET (GPA_OPTION_MENU (ps->trmenu)->menu));
 	gtk_label_set_mnemonic_widget ((GtkLabel *) l, GPA_OPTION_MENU (ps->trmenu)->menu);
 
 	/* Preview frame */
