@@ -25,17 +25,31 @@
  */
 
 #include <config.h>
+#include "gnome-print-preview.h"
 
 #include <string.h>
 #include <math.h>
 #include <gdk-pixbuf/gdk-pixbuf.h>
 
 #include <libgnomecanvas/gnome-canvas-clipgroup.h>
+
+#include <libgnomeprint/private/gnome-print-private.h>
 #include <libgnomeprint/private/gp-gc-private.h>
 #include <libgnomeprint/private/gnome-glyphlist-private.h>
 
 #include "gnome-canvas-hacktext.h"
-#include "gnome-print-preview-private.h"
+
+typedef struct _GnomePrintPreviewPrivate GnomePrintPreviewPrivate;
+
+struct _GnomePrintPreview {
+	GnomePrintContext pc;
+
+	GnomePrintPreviewPrivate *priv;
+};
+
+struct _GnomePrintPreviewClass {
+	GnomePrintContextClass parent_class;
+};
 
 struct _GnomePrintPreviewPrivate {
 	GPGC * gc;
@@ -55,6 +69,12 @@ struct _GnomePrintPreviewPrivate {
 
 static GnomePrintContextClass *parent_class;
 
+#define GPP_COLOR_RGBA(color, ALPHA) \
+	((guint32) (ALPHA              | \
+	 (((color).red   / 256) << 24) | \
+	 (((color).green / 256) << 16) | \
+	 (((color).blue  / 256) << 8)))
+
 static void
 outline_set_style_cb (GtkWidget *canvas, GnomeCanvasItem *item)
 {
@@ -315,6 +335,7 @@ glyphlist_set_style_cb (GtkWidget *canvas, GnomeCanvasItem *item)
 		}
 	}
 	gnome_canvas_item_set (item, "glyphlist", new, NULL);
+	gnome_glyphlist_unref (new);
 }
 
 static int
